function [expectationVectorOut , stdOut , Theta1] = computeFeatures_MultiDim(data , expectationVector , Theta0 , cdtParams)
%
% function [expectationVectorOut , stdOut] = computeFeatures_MultiDim(data , expectationVector , Theta0, cdtParams)
%
% compute the expected feature value including the incoming data
%
% input description
%
% data                  incoming observation.
%                       size(data) = [dimension of observation (d in [1]), samples of each window (nu in[1]), observation seen so far (T in [1])];
% expectationVector     vector/matrix containing the previous estimates of feature expectations
% Theta0                structure describing features in stationary contitions, obtained from computeTrainFeatures_MultiDim
% weights               The weights used in the polynomial estimator ('uniform' [default], 'parabolic' or 'constant+alpha');
%
%
% output
%
% expectationVectorOut  value of the polynomial estimator
% stdOut                standard deviation of the polynomial estimator
% Theta1                updated parameters (only when higher order interpolation is used).
%                       when other values of weights are set, Theta1 = Theta0
%                       fields concerning features computed from training set, Theta1.stdDev are NEVER updated
%
% Giacomo Boracchi
% Politecnico di Milano
% January 2010
% giacomo.boracchi@polimi.it
%
% Revision History
% October 2010 -   added LPA estimates for both features
% December 2010 -  added 1st order interpolation for sample mean
%
% please cite this work as
% [1] Cesare Alippi, Giacomo Boracchi and Manuel Roveri,
% "Change Detection Tests Using the ICI rule"
% in Proceedings of IJCNN 2010, the International Joint Conference on Neural Networks 18 - 23 July, 2010 Barcelona, Spain.
%
% [2] Cesare Alippi, Giacomo Boracchi and Manuel Roveri,
% "A just-in-time adaptive classification system based on the intersection of confidence intervals rule"
%  Neural Networks, Elsevier vol. 24 (2011), pp. 791-800 (doi: 10.1016/j.neunet.2011.05.012)
%
% References
% [Box64] Box, George EP and Cox, David R
% "An analysis of transformations",
% Journal of the Royal Statistical Society. Series B (Methodological)
% pp 211--252 1964,
%
% [Mudholkar81] Mudholkar G. S., Trivedi Mom. C.: A Gaussian Approximation to the Distribution of the
% Sample Variance for Nonnormal Populations, Journal of the American Statistical
% Association, Vol. 76, No. 374 (Jun., 1981), pp. 479-485

weights = cdtParams.weights;

% it is modified only when 1st order interpolation is used
Theta1 = Theta0;
    
%% compute the mean

% il numero di colonne di expectation vector rappresenta il numero di features values osservati finora
nPreviousElements = size(expectationVector , 2);

% faccio la media sulla terza dimensione dello slot contenente gli ultimi samples da analizzare
current_expectation_estimate = mean(data(: , nPreviousElements + 1 , :) , 3);

% apply the powerlaw trasnform to make data approximatively Gaussian distributed [Box64]
if (strcmpi(weights , 'pointwise')) % for training set it is the same
    if (strcmpi(cdtParams.pointwise.transformType, 'boxcox') || strcmpi(cdtParams.pointwise.transformType, 'manly'))
      current_expectation_estimate = Theta0(1).transform(current_expectation_estimate);
    end
end

% iterative expression to compute the expectation over sample mean. It is the same for Bernoulli samples.
if (strcmpi(weights , 'uniform') || strcmpi(weights , 'Bernulli') || strcmpi(weights , 'Bernoulli') || strcmpi(weights , 'pointwise'))
    mu_hat = (expectationVector(: , end , 1) * nPreviousElements + current_expectation_estimate) / (nPreviousElements + 1);
    stdEst = Theta0(1).stdDev ./ sqrt(nPreviousElements + 1);
end

if (strcmpi(weights , 'parabolic'))
    means = mean(data(: , 1 : nPreviousElements + 1 , :) , 3);
    [mu_hat , stdEst] = zeroOrderFitParabolicWeights(means, Theta0(1).stdDev);
end

if (strcmpi(weights , 'constant+alpha')) % for training set it is the same
    warning('Not implemented constant+alpha case yet')
    return;
    %     [mu_hat , sigmaEst] = zeroOrderFitConstant_Alpha_Weights(data , sigma_noise , );
end

if (strcmpi(weights , 'order1'))
    % point where the estimate is provided, we take the left-point = 0;
    [mu_hat , Theta1] = firstOrderFit(current_expectation_estimate , Theta0);
    mu_hat = Theta1.mu_hat;
    stdEst = Theta1.stdEst;
end

% update vectors
expectationVectorOut(: , : , 1) = mu_hat;
stdOut(: , : , 1) = stdEst;

%% compute the sample variance (only when Theta0 has two dimensions)

if(numel(Theta0) > 1)
    % perform detrending before computing sample variance, then use 0 order interpolation
    if (strcmpi(weights , 'order1'))
        detrendingFilter = [-1 1];
        vals = reshape(data(: , nPreviousElements + 1 , :) , [1 , length(data(: , nPreviousElements + 1 , :))]);
        detrendedVals = convn(vals , detrendingFilter , 'same');
        % downsampling sempre presente
        S2 = var(detrendedVals(1 : 2 : end));
    else
        S2 = mean((data(: , nPreviousElements + 1 , :) - repmat(current_expectation_estimate , [1 , 1 , size(data , 3)])) .^ 2 , 3);
    end
    
    if size(data , 1) == 1
        % compute the Gaussian approximating function
        T = Theta0(2).transform(S2);
    elseif size(data , 1) == 2
        % compute the Gaussian approximating function
        T = Theta0(2).transform(S2(1) , S2(2));
    else
        for dd = 1 : size(data , 1)
            T(dd) = (S2(dd) ./ Theta0(2).k1(dd)) .^ Theta0(2).h0(dd);
        end
    end
    
    if (strcmpi(weights , 'uniform') || strcmpi(weights , 'order1'))
        mu_hat = (expectationVector(: , end , 2) * nPreviousElements + T) / (nPreviousElements + 1);
        stdEst = Theta0(2).stdDev ./ sqrt(nPreviousElements + 1);
    end
    
    if (strcmpi(weights , 'parabolic'))
        values = [expectationVector(: , 1 : end , 2) , T];
        [mu_hat , stdEst] = zeroOrderFitParabolicWeights(values , Theta0(2).stdDev);
    end
    
    expectationVectorOut(: , : , 2) = mu_hat;
    stdOut(: , : , 2) = stdEst;
end

% % expectationVector ha queste dimensioni
% %
% % size (ExpectationVector) = sampleDimension x #stimeCampioni x numberOfFeatures
%
% % il numero di colonne
% nPreviousElements=size(expectationVector,2);
% for tt=1:size(expectationVector,3)
%     % l'ultimo elemento in expectationVector  la media di tutti i nPreviousElements
%     expectationVectorOut(:,:,tt) = (expectationVector(: , end , tt) * nPreviousElements + Val(: , tt)) / (nPreviousElements + 1);
% end

