function  [refined_val , testDetectingChange , midpoint] = f_ICI_refine_test(DataSetComplete, detected, dimensionDetected, tsLength, testParams, Theta0)
%
% function  [refined , testDetectingChange , midpoint] = f_ICI_refine_test(DataSetComplete, detected, dimensionDetected , tsLength , testParams, Theta0)
%
% Implement the refinement procedure presented in 
%
% [1] Cesare Alippi, Giacomo Boracchi and Manuel Roveri,
% "A just-in-time adaptive classification system based on the intersection of confidence intervals rule"
%  Neural Networks, Elsevier vol. 24 (2011), pp. 791-800 (doi: 10.1016/j.neunet.2011.05.012)
%
% [2] Cesare Alippi, Giacomo Boracchi and Manuel Roveri,
% "Adaptive Classifiers with ICI-based Adaptive Knowledge Base Management"
% ICANN 2010, 20th International Conference on Artificial Neural Networks, September 15-18, 2010, Thessaloniki, Greece,
% Lecture Notes on Computer Science. Springer Berlin / Heidelberg, vol 6353, pg 458-467
%
% Revision History
% April 2010              first release
% December 2013     added Theta0 as an optional parameter to avoid multiple calls to computeTrainingFeatures()
%
% Giacomo Boracchi
% Politecnico di Milano
% giacomo.boracchi@polimi.it

% if it does not exist passes an empty variable -> implies call to computeTrainingFeatures
if (~ exist('Theta0' , 'var'))
    Theta0 = [];
end

%% 
K = testParams.K_refinement;
windowWidth = testParams.windowWidth;

% the test is meant for scalar data, wheras the test was used separately on each dimension of the dataset, we consider here only the dimension where the cahange was detected
if size(DataSetComplete , 1) > 1
    DataSet = DataSetComplete(dimensionDetected , 1 : detected);
else
    DataSet = DataSetComplete(1 : detected);
end

split = 1;
s = 1;
% la lunghezza  pari alla met del validation set (lunghezza DataSet - trainingSet)

% take the first midpoint
midpoint = floor((length(DataSet) - tsLength) / K) + tsLength;

% initialize the refinement in case there won't be any detection of the dataset
refined = detected;
refined_tmp = [];

while (split)
	% define the dataset to be analyzed
    ValidationSet_temp = DataSet(midpoint(s) + 1: end);
    roundedValidationSetLength = floor(length(ValidationSet_temp) / windowWidth) * windowWidth;
    ValidationSet = ValidationSet_temp(end - roundedValidationSetLength +1 : end);
    TrainingSet = DataSet(1 : tsLength);
    DataSetReduced = [TrainingSet , ValidationSet];
    
	% this is used only for cdt on linear trends
    if strcmpi(testParams.weights , 'order1')
        % modifico il DataSet per corregere i trend
        % devo stimare il valore della retta di regressione nel midpoint
        validationSetStartsAt = length(DataSet) - length(ValidationSet);
        regParams = polyfit([0 : validationSetStartsAt - 1] , DataSet(1 : validationSetStartsAt) , 1);
        midpointVal = regParams(1) * (validationSetStartsAt - 1) + regParams(2);
        tsVal =  regParams(1) * (tsLength - 1) + regParams(2);
        % li riallineo in modo da evitare abrupt changes
        ValidationSet = ValidationSet - midpointVal + tsVal;
        DataSetReduced = [TrainingSet , ValidationSet];
    end

	% run the CDT on th reduced dataset, specifying the GammaRefinement parameter to be used
    [refined_tmp(s) , testDetectingChange(s) , Theta0 , dimensionWhereChangeDetected(s)] = f_ICI_test_MultiDim(DataSetReduced, tsLength, testParams, testParams.GammaRefinement, testParams.numberOfFeatures, Theta0);
    
    % When the refinement does not provide any detection we trust the previous one, which probably was close enough
    if refined_tmp(s) == 0
        % no change detected in the refinement procedure -> set T_hat
        refined(s) = detected;
        
    else
        % update refined position
        refined(s) = midpoint(s) + refined_tmp(s) - tsLength;
	end
    
    s = s + 1;
    
    % update the midpoint position
    midpoint(s) = floor(length(ValidationSet) /  K) + midpoint(s-1);
    
    % Check if the split has not reached the end of the dataset or if the next split point has exceeded one of the estimates obtained during the refinement procedure
    if(min(refined) <= midpoint(s) || midpoint(s) == midpoint(s-1))
        split = 0;
    end
    
end

% return value
refined_val = min(refined);