% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Tutorial "Learning in Non-Stationary Environments" - WCCI 2016
% Prof. Manuel Roveri
% Politecnico di Milano, Italy
% Jul, 2016
%
% ----- "Learning in Nonstationary Environments"
% -> Passive and Active Approaches
% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

clear;
clc;
close all;
addpath('./functions');

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%       EXPERIMENT PARAMETERS
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Number of data batches considered by the experiment
Niter = 10;

% Number of samples for each batch
N = 50;

% Number of NN hidden nodes
n_hidden_nodes = 10;

%Noise standard deviation
std_dev = 0.05;

%Characterization of the change: 'abrupt' or 'drift'
CHANGE_TYPE = 'abrupt'; %'abrupt' or 'drift'
CHANGE_PARAMETER = 2; % Magnitude of the change 
CHANGE_ITERATION = 5; % The number of the batch the change started

%LNSE solution: "NO_Adaptation', 'ACTIVE', 'PASSIVE'
ADAPTIVE_SOLUTION = 'NO_Adaptation';
THRESHOLD_ACTIVE_CDT = 10; % The CDT of the active solution is implemented through a threshold on the residual. The value of this threshold is THRESHOLD_ACTIVE_CDT


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%       MAIN
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
tstart = tic;

fprintf('----- BEGIN -------\n\n');
fprintf('MATLAB Code: TUTORIAL LNSE - WCCI 2016 \n\n');
fprintf('Solution: %s\n',ADAPTIVE_SOLUTION);
fprintf('Change Type: %s\n',CHANGE_TYPE);

% Part 1) Training of the NN %%%%%%%%%%%

%Training data generation
Xt = 4*rand(1,N)-2; %uniform samples on [-2;2]
noise = std_dev * randn(1,N);
Yt = nonlin_fun(Xt,noise);

%%%%%%% Network training
% Initializing network
neural_net = fitnet(n_hidden_nodes);
%Training network
neural_net = train(neural_net,Xt,Yt);


%% Plotting results
plot_nn_regression(@nonlin_fun,neural_net,Xt,Yt);
title('Original Regression Problem');


bias = 0;

fprintf('Testing phase ... \n'); 
for niter = 1:Niter
    
    %% Generating new data
    X_alt = 4*rand(1,N)-2; %uniform samples on [-2;2]
    if niter > CHANGE_ITERATION
        switch lower(CHANGE_TYPE)
            case {'abrupt'}
                bias = CHANGE_PARAMETER;
            case {'drift'}
                bias = (niter-CHANGE_ITERATION) * CHANGE_PARAMETER;
                bias_vector (niter) = bias;
            otherwise
                bias = 0;
        end
    end
    noise = std_dev * randn(1,N);
    Y_alt = nonlin_fun_biased(X_alt,noise,bias);
    
    % Plotting results
    figure(niter+1);
    plot(X_alt,Y_alt,'bd');
    hold on;
    Y_est = neural_net(X_alt);
    plot(X_alt,Y_est,'rx');
    if (niter < CHANGE_ITERATION) || (strcmp(CHANGE_TYPE,'abrupt')==0 && strcmp(CHANGE_TYPE,'drift')==0)
        title(['Batch #' num2str(niter) ': NO change in the regression problme ']);
    else
        title(['Batch #' num2str(niter) ': '  CHANGE_TYPE ' CHANGE started at ' num2str(CHANGE_ITERATION)]);
    end
    
    error(niter) = sqrt(mean((Y_alt - Y_est).^2)); % EVALUATE THE BIAS
    
    switch lower(ADAPTIVE_SOLUTION)
        case {'passive'}
            %Training Neural Network
            neural_net = train(neural_net,X_alt,Y_alt);
            fprintf('Passive Approach: Re-training of the NN (iteration %d) \n',niter);
        case {'active'}
            if error(niter) > THRESHOLD_ACTIVE_CDT
                neural_net = train(neural_net,X_alt,Y_alt);
                fprintf('Active Approach: Re-training of the NN (iteration %d) \n',niter);
            end
            
    end
    
    %% New training
    %neural_net_alt = train(neural_net,X_alt,Y_alt);
    
    %% Plotting results
    %plot_nn_regression(@nonlin_fun_biased,neural_net_alt,X_alt,Y_alt);
    %title('Regression Problem after Adaptation');
end

telapsed = toc(tstart);
fprintf('Elapsed time (s): %.1f\n',telapsed);

figure, plot(error,'-*');
xlabel('Batch');
ylabel('RMSE');
title('Evolution of the RMSE over the batches');
fprintf('----- END -------\n\n');